"""Custom template tags for Mork."""

from jinja2_simple_tags import StandaloneTag

from mork.conf import settings
from mork.utils import svg_to_datauri


class SVGStaticTag(StandaloneTag):
    """Extension Jinja tag for converting SVG files to data URI."""

    tags = {"svg_static"}

    def render(self, path: str):
        """Return a SVG static file into data URI format."""
        full_path = settings.STATIC_PATH / path
        if full_path.exists():
            return svg_to_datauri(full_path)
        return ""



"""Mork models."""

from enum import Enum, unique

from pydantic import BaseModel

from mork.celery.tasks import delete_inactive_users, warn_inactive_users


@unique
class TaskStatus(str, Enum):
    """Task statuses."""

    FAILURE = "FAILURE"
    PENDING = "PENDING"
    RECEIVED = "RECEIVED"
    RETRY = "RETRY"
    REVOKED = "REVOKED"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"


@unique
class TaskType(str, Enum):
    """Possible task types."""

    EMAILING = "emailing"
    DELETION = "deletion"


class TaskCreate(BaseModel):
    """Model for creating a new task."""

    type: TaskType


class TaskResponse(BaseModel):
    """Model for a task response."""

    id: str
    status: TaskStatus


TASK_TYPE_TO_FUNC = {
    TaskType.EMAILING: warn_inactive_users,
    TaskType.DELETION: delete_inactive_users,
}



"""Tests for authentication for the Mork API."""

import pytest
from fastapi import HTTPException, status

from mork.api.auth import authenticate_api_key
from mork.conf import settings


@pytest.mark.anyio
async def test_authenticate_api_key():
    """Test the authentication with a wrong API key."""
    settings.API_KEYS = ["valid_key_1", "valid_key_2"]

    # Test with a valid API key
    assert authenticate_api_key("valid_key_1") is None
    assert authenticate_api_key("valid_key_2") is None

    # Test with an invalid API key
    with pytest.raises(HTTPException) as exc_info:
        authenticate_api_key("invalid_key")

    assert exc_info.value.status_code == status.HTTP_401_UNAUTHORIZED
    assert exc_info.value.detail == "Missing or invalid API key"



"""Factory classes for generating fake data for testing."""

import factory

from mork import models


class EmailStatusFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory for generating fake email status entries."""

    class Meta:
        """Factory configuration."""

        model = models.EmailStatus

    email = factory.Faker("email")
    sent_date = factory.Faker("date_time")



"""Email related functions."""

import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from logging import getLogger
from smtplib import SMTPException

from jinja2 import Environment, FileSystemLoader

from mork.conf import settings
from mork.exceptions import EmailSendError
from mork.templatetags.extra_tags import SVGStaticTag

logger = getLogger(__name__)


def render_template(template: str, context) -> str:
    """Render a Jinja template into HTML."""
    template_env = Environment(
        loader=FileSystemLoader(
            [
                settings.ROOT_PATH / "templates/html",
                settings.ROOT_PATH / "templates/text",
            ]
        ),
        autoescape=True,
        extensions=[SVGStaticTag],
    )
    template = template_env.get_template(template)
    return template.render(**context)


def send_email(email_address: str, username: str):
    """Initialize connection to SMTP and send a warning email."""
    template_vars = {
        "title": "Votre compte va être supprimé dans 30 jours.",
        "email": email_address,
        "fullname": username,
        "site": {
            "name": settings.EMAIL_SITE_NAME,
            "url": settings.EMAIL_SITE_BASE_URL,
            "login_url": settings.EMAIL_SITE_LOGIN_URL,
        },
    }
    html = render_template(
        "warning_email.html",
        template_vars,
    )

    text = render_template(
        "warning_email.txt",
        template_vars,
    )

    # Create a multipart message (with MIME type multipart/alternative) and set headers
    message = MIMEMultipart("alternative")
    message["From"] = settings.EMAIL_FROM
    message["To"] = email_address
    message["Subject"] = "Votre compte va bientôt être supprimé"

    # Attach the HTML parts. According to RFC 2046, the last part of a multipart
    # message, in this case the HTML message, is best and preferred
    message.attach(MIMEText(text, "plain"))
    message.attach(MIMEText(html, "html"))

    # Send the email
    with smtplib.SMTP(
        host=settings.EMAIL_HOST, port=settings.EMAIL_PORT
    ) as smtp_server:
        if settings.EMAIL_USE_TLS:
            smtp_server.starttls()
        if settings.EMAIL_HOST_USER and settings.EMAIL_HOST_PASSWORD:
            smtp_server.login(
                user=settings.EMAIL_HOST_USER,
                password=settings.EMAIL_HOST_PASSWORD,
            )
        try:
            smtp_server.sendmail(
                from_addr=settings.EMAIL_FROM,
                to_addrs=email_address,
                msg=message.as_string(),
            )
        except SMTPException as exc:
            logger.error(f"Sending email failed: {exc} ")
            raise EmailSendError("Failed sending an email") from exc



"""Mork models."""

import datetime
import uuid

from sqlalchemy import DateTime, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    """Base class for all models in the database."""

    filtered_attrs = []

    def safe_dict(self):
        """Return a dictionary representation of the model."""
        return {
            c.name: getattr(self, c.name)
            for c in self.__table__.columns
            if c.name not in self.filtered_attrs
        }


class EmailStatus(Base):
    """Model for storing email statuses."""

    __tablename__ = "email_status"

    filtered_attrs = ["email"]

    id: Mapped[int] = mapped_column(
        UUID(as_uuid=True), primary_key=True, default=uuid.uuid4
    )
    email: Mapped[str] = mapped_column(String(254), unique=True)
    sent_date: Mapped[datetime.datetime] = mapped_column(DateTime)



"""API routes related to application health checking."""

import logging

from fastapi import APIRouter

logger = logging.getLogger(__name__)

router = APIRouter()


@router.get("/__lbheartbeat__")
async def lbheartbeat() -> None:
    """Load balancer heartbeat.

    Return a 200 when the server is running.
    """
    return



"""Utility functions."""

from pathlib import Path

from datauri import DataURI


def svg_to_datauri(path: Path | str):
    """Return the data URI string of an SVG image."""
    return str(DataURI.from_file(path))



"""Mork database connection."""

import logging

from sqlalchemy import NullPool, create_engine
from sqlalchemy.orm import Session

from mork.conf import settings

logger = logging.getLogger(__name__)


class MorkDB:
    """Class to connect to the Mork database."""

    session = None

    def __init__(self):
        """Initialize SqlAlchemy engine and session."""
        # Disable pooling as SQLAlchemy connections cannot be shared accross processes,
        # and Celery forks processes by default
        self.engine = create_engine(
            settings.DB_URL, echo=settings.DB_DEBUG, poolclass=NullPool
        )
        self.session = Session(self.engine)



"""Main module for Mork API."""

from functools import lru_cache
from typing import Dict, List, Union
from urllib.parse import urlparse

import sentry_sdk
from fastapi import FastAPI

from mork import __version__
from mork.api.routes import health, tasks
from mork.conf import settings


@lru_cache(maxsize=None)
def get_health_check_routes() -> List:
    """Return the health check routes."""
    return [route.path for route in health.router.routes]


def filter_transactions(event: Dict, hint) -> Union[Dict, None]:  # noqa: ARG001
    """Filter transactions for Sentry."""
    url = urlparse(event["request"]["url"])

    if settings.SENTRY_IGNORE_HEALTH_CHECKS and url.path in get_health_check_routes():
        return None

    return event


if settings.SENTRY_DSN is not None:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        traces_sample_rate=settings.SENTRY_API_TRACES_SAMPLE_RATE,
        release=__version__,
        environment=settings.SENTRY_EXECUTION_ENVIRONMENT,
        max_breadcrumbs=50,
        before_send_transaction=filter_transactions,
    )

app = FastAPI()
app.include_router(health.router)
app.include_router(tasks.router)



"""Tests of the Mork models."""

from mork.factories import EmailStatusFactory


def test_models_user_safe_dict(db_session):
    """Test the `safe_dict` method for the EmailStatus model."""
    email_status = EmailStatusFactory()

    assert email_status.safe_dict() == {
        "id": email_status.id,
        "sent_date": email_status.sent_date,
    }



"""Fixtures for the mork app."""

from typing import AsyncIterator

import pytest
from httpx import AsyncClient

from mork.api import app
from mork.conf import settings


@pytest.fixture
@pytest.mark.anyio
async def http_client() -> AsyncIterator[AsyncClient]:
    """Handle application lifespan while yielding asynchronous HTTP client."""
    async with AsyncClient(app=app, base_url=settings.SERVER_URL) as client:
        yield client



"""Exceptions for Mork."""


class EmailAlreadySent(Exception):
    """Raised when an email has already been sent to this user."""


class EmailSendError(Exception):
    """Raised when an error occurs when sending an email."""



"""Initialize database

Revision ID: 608d075c6e99
Revises: 
Create Date: 2024-08-02 08:33:18.791338

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "608d075c6e99"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "email_status",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("email", sa.String(length=254), nullable=False),
        sa.Column("sent_date", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("email_status")
    # ### end Alembic commands ###



"""Tests for Mork Celery tasks."""

from mork.celery.tasks import check_email_already_sent, mark_email_status
from mork.factories import EmailStatusFactory


def test_check_email_already_sent(monkeypatch, db_session):
    """Test the `check_email_already_sent` function."""
    email_address = "test_email@example.com"

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.MorkDB", MockMorkDB)
    EmailStatusFactory.create_batch(3)

    assert not check_email_already_sent(email_address)

    EmailStatusFactory.create(email=email_address)
    assert check_email_already_sent(email_address)


def test_mark_email_status(monkeypatch, db_session):
    """Test the `mark_email_status` function."""

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.MorkDB", MockMorkDB)

    # Write new email status entry
    new_email = "test_email@example.com"
    mark_email_status(new_email)
    assert check_email_already_sent(new_email)



"""Factory classes for generating fake data for testing."""

import factory
from faker import Faker
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from mork.edx import models
from mork.edx.models import Base

faker = Faker()
engine = create_engine("sqlite+pysqlite:///:memory:", echo=False, pool_pre_ping=True)
session = Session(engine)
Base.metadata.create_all(engine)


class EdxUserFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory for generating fake Open edX users."""

    class Meta:
        """Factory configuration."""

        model = models.User
        sqlalchemy_session = session

    id = factory.Sequence(lambda n: n + 1)
    username = factory.Sequence(lambda n: f"{faker.user_name()}{n}")
    email = factory.Faker("email")
    is_staff = False
    is_superuser = False
    last_login = factory.Faker("date_time")



"""Configurations for Mork."""

import io
from pathlib import Path
from typing import Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Pydantic model for Mork's global environment & configuration settings."""

    model_config = SettingsConfigDict(
        case_sensitive=True,
        env_file=".env",
        env_file_encoding=getattr(io, "LOCALE_ENCODING", "utf-8"),
        env_nested_delimiter="__",
        env_prefix="MORK_",
        extra="ignore",
    )

    # Mork server
    API_SERVER_PROTOCOL: str = "http"
    API_SERVER_HOST: str = "localhost"
    API_SERVER_PORT: int = 8100
    API_KEYS: list[str] = ["APIKeyToBeChanged"]

    # API Root path
    # (used at least by everything that is alembic-configuration-related)
    ROOT_PATH: Path = Path(__file__).parent

    # Alembic
    ALEMBIC_CFG_PATH: Path = ROOT_PATH / "alembic.ini"

    # Static path
    STATIC_PATH: Path = ROOT_PATH / "static"

    # Mork database
    DB_ENGINE: str = "postgresql+psycopg"
    DB_HOST: str = "postgresql"
    DB_NAME: str = "mork-db"
    DB_USER: str = "fun"
    DB_PASSWORD: str = "pass"
    DB_PORT: int = 5432
    DB_DEBUG: bool = False
    TEST_DB_NAME: str = "test-mork-db"

    # EDX database
    EDX_DB_ENGINE: str = "mysql+pymysql"
    EDX_DB_HOST: str = "mysql"
    EDX_DB_NAME: str = "edxapp"
    EDX_DB_USER: str = "edxapp"
    EDX_DB_PASSWORD: str = "password"
    EDX_DB_PORT: int = 3306
    EDX_DB_DEBUG: bool = False

    # Emails
    EMAIL_HOST: str = "mailcatcher"
    EMAIL_HOST_USER: str = ""
    EMAIL_HOST_PASSWORD: str = ""
    EMAIL_PORT: int = 1025
    EMAIL_USE_TLS: bool = False
    EMAIL_FROM: str = ""
    EMAIL_RATE_LIMIT: str = "100/m"
    EMAIL_MAX_RETRIES: int = 3
    EMAIL_SITE_NAME: str = ""
    EMAIL_SITE_BASE_URL: str = ""
    EMAIL_SITE_LOGIN_URL: str = ""

    # Celery
    broker_url: str = Field("redis://redis:6379/0", alias="MORK_CELERY_BROKER_URL")
    result_backend: str = Field(
        "redis://redis:6379/0", alias="MORK_CELERY_RESULT_BACKEND"
    )
    broker_transport_options: dict = Field(
        {}, alias="MORK_CELERY_BROKER_TRANSPORT_OPTIONS"
    )
    task_default_queue: str = Field("celery", alias="MORK_CELERY_TASK_DEFAULT_QUEUE")

    # Sentry
    SENTRY_DSN: Optional[str] = None
    SENTRY_EXECUTION_ENVIRONMENT: str = "development"
    SENTRY_API_TRACES_SAMPLE_RATE: float = 1.0
    SENTRY_IGNORE_HEALTH_CHECKS: bool = False

    @property
    def DB_URL(self) -> str:
        """Get the Mork database URL as required by SQLAlchemy."""
        return (
            f"{self.DB_ENGINE}://"
            f"{self.DB_USER}:{self.DB_PASSWORD}@"
            f"{self.DB_HOST}/{self.DB_NAME}"
        )

    @property
    def TEST_DB_URL(self) -> str:
        """Get the test database URL as required by SQLAlchemy."""
        return (
            f"{self.DB_ENGINE}://"
            f"{self.DB_USER}:{self.DB_PASSWORD}@"
            f"{self.DB_HOST}/{self.TEST_DB_NAME}"
        )

    @property
    def EDX_DB_URL(self) -> str:
        """Get the edx database URL as required by SQLAlchemy."""
        return (
            f"{self.EDX_DB_ENGINE}://"
            f"{self.EDX_DB_USER}:{self.EDX_DB_PASSWORD}@"
            f"{self.EDX_DB_HOST}/{self.EDX_DB_NAME}"
        )

    @property
    def SERVER_URL(self) -> str:
        """Get the full server URL."""
        return f"{self.API_SERVER_PROTOCOL}://{self.API_SERVER_HOST}:{self.API_SERVER_PORT}"


settings = Settings()



"""Mork celery configuration file."""

from celery import Celery

app = Celery("mork", include=["mork.celery.tasks"])

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("mork.conf:settings", namespace="CELERY")



"""Tests for the Mork API '/tasks/' endpoints."""

from unittest.mock import Mock, patch

import pytest
from httpx import AsyncClient


@pytest.mark.anyio
async def test_tasks_auth(http_client: AsyncClient):
    """Test required authentication for tasks endpoints."""
    # FastAPI returns a 403 error (instead of a 401 error) if no API token is given
    # see https://github.com/tiangolo/fastapi/discussions/9130
    assert (await http_client.post("/tasks/")).status_code == 403
    assert (await http_client.options("/tasks/")).status_code == 403
    assert (await http_client.get("/tasks/status/1234")).status_code == 403


@pytest.mark.anyio
@pytest.mark.parametrize("task_type", ["emailing", "deletion"])
async def test_create_task(
    http_client: AsyncClient, auth_headers: dict, task_type: str
):
    """Test creating a task with valid data."""

    mock_task_create = {"type": task_type}

    celery_task = Mock()
    celery_task.delay.return_value.task_id = "1234"

    with patch.dict("mork.api.tasks.TASK_TYPE_TO_FUNC", {task_type: celery_task}):
        response = await http_client.post(
            "/tasks/", headers=auth_headers, json=mock_task_create
        )
        response_data = response.json()

        assert response.status_code == 202
        assert response_data.get("id")
        assert response_data.get("status") == "PENDING"
        assert (
            response.headers["location"] == f"/tasks/status/{response_data.get("id")}"
        )
        celery_task.delay.assert_called()


@pytest.mark.anyio
async def test_create_task_invalid_type(http_client: AsyncClient, auth_headers: dict):
    """Test creating a task with valid data."""

    # Without a type
    with patch.dict("mork.api.tasks.TASK_TYPE_TO_FUNC"):
        response = await http_client.post("/tasks/", headers=auth_headers)

        assert response.status_code == 422

    # With a wrong type
    mock_task_create = {"type": "wrong_type"}

    celery_task = Mock()
    celery_task.delay.return_value.task_id = "1234"

    with patch.dict("mork.api.tasks.TASK_TYPE_TO_FUNC"):
        response = await http_client.post(
            "/tasks/", headers=auth_headers, json=mock_task_create
        )

        assert response.status_code == 422


@pytest.mark.anyio
async def test_get_available_tasks(http_client: AsyncClient, auth_headers: dict):
    """Test getting available tasks."""
    response = await http_client.options("/tasks/", headers=auth_headers)
    response_data = response.json()
    assert response.status_code == 200
    assert response.headers["allow"] == "POST"
    assert sorted(response_data.get("task_types")) == ["deletion", "emailing"]


@pytest.mark.anyio
async def test_get_task_status(http_client: AsyncClient, auth_headers: dict):
    """Test getting task status."""

    task_id = "1234"
    celery_result = Mock(task_id)
    celery_result(task_id).state = "SUCCESS"

    with patch("mork.api.tasks.AsyncResult", celery_result):
        response = await http_client.get(
            f"/tasks/status/{task_id}",
            headers=auth_headers,
        )
        response_data = response.json()
        assert response.status_code == 200
        assert response_data.get("status") == "SUCCESS"



"""Tests for Mork mail functions."""

import smtplib
from unittest.mock import MagicMock

import pytest

from mork.exceptions import EmailSendError
from mork.mail import render_template, send_email


def test_render_template():
    """Test the `render_template` function."""
    template_vars = {
        "title": "Votre compte va être supprimé dans 30 jours.",
        "email": "test@example.com",
        "fullname": "John Doe",
        "site": {
            "name": "Example site",
            "url": "http://base_url.com",
            "login_url": "http://url.com/login",
        },
    }
    render_html = render_template("warning_email.html", template_vars)
    assert template_vars["title"] in render_html
    assert template_vars["email"] in render_html
    assert template_vars["fullname"] in render_html
    assert template_vars["site"]["name"] in render_html
    assert template_vars["site"]["url"] in render_html
    assert template_vars["site"]["login_url"] in render_html
    assert 'src="data:' in render_html

    render_text = render_template("warning_email.txt", template_vars)
    assert template_vars["title"] in render_text
    assert template_vars["email"] in render_text
    assert template_vars["fullname"] in render_text
    assert template_vars["site"]["name"] in render_text
    assert template_vars["site"]["url"] in render_text
    assert template_vars["site"]["login_url"] in render_text
    assert "data:" in render_text


def test_send_email(monkeypatch):
    """Test the `send_email` function."""

    mock_SMTP = MagicMock()
    monkeypatch.setattr("mork.mail.smtplib.SMTP", mock_SMTP)

    test_address = "john.doe@example.com"
    test_username = "JohnDoe"
    send_email(email_address=test_address, username=test_username)

    assert mock_SMTP.return_value.__enter__.return_value.sendmail.call_count == 1


def test_send_email_with_smtp_exception(monkeypatch):
    """Test the `send_email` function with an SMTP exception."""

    mock_SMTP = MagicMock()
    mock_SMTP.return_value.__enter__.return_value.sendmail.side_effect = (
        smtplib.SMTPException
    )

    monkeypatch.setattr("mork.mail.smtplib.SMTP", mock_SMTP)

    test_address = "john.doe@example.com"
    test_username = "JohnDoe"

    with pytest.raises(EmailSendError, match="Failed sending an email"):
        send_email(email_address=test_address, username=test_username)



"""Fixtures for the authorization headers of Mork api."""

import pytest

from mork.conf import settings


@pytest.fixture
def auth_headers() -> dict:
    """Generate authentication headers with an API token.

    Returns:
        dict: A dictionary containing the 'X-API-Key' header with an API key.

    Example:
        To use this fixture in a test function, you can do the following::

            def test_authenticated_request(client):
                headers = auth_headers()
                response = client.get("/protected/resource", headers=headers)
                assert response.status_code == 200

    """
    return {"X-API-Key": f"{settings.API_KEYS[0]}"}



"""Mork Celery tasks."""

from datetime import datetime
from logging import getLogger

from sqlalchemy import select

from mork.celery.celery_app import app
from mork.conf import settings
from mork.database import MorkDB
from mork.exceptions import EmailAlreadySent, EmailSendError
from mork.mail import send_email
from mork.models import EmailStatus

logger = getLogger(__name__)


@app.task
def warn_inactive_users():
    """Celery task to warn inactive users by email."""
    pass


@app.task
def delete_inactive_users():
    """Celery task to delete inactive users accounts."""
    pass


@app.task(
    bind=True,
    rate_limit=settings.EMAIL_RATE_LIMIT,
    retry_kwargs={"max_retries": settings.EMAIL_MAX_RETRIES},
)
def send_email_task(self, email_address: str, username: str):
    """Celery task that sends an email to the specified user."""
    # Check that user has not already received a warning email
    if check_email_already_sent(email_address):
        raise EmailAlreadySent("An email has already been sent to this user")

    try:
        send_email(email_address, username)
    except EmailSendError as exc:
        logger.exception(exc)
        raise self.retry(exc=exc) from exc

    # Write flag that email was correctly sent to this user
    mark_email_status(email_address)


def check_email_already_sent(email_adress: str):
    """Check if an email has already been sent to the user."""
    db = MorkDB()
    query = select(EmailStatus.email).where(EmailStatus.email == email_adress)
    result = db.session.execute(query).scalars().first()
    db.session.close()
    return result


def mark_email_status(email_address: str):
    """Mark the email status in database."""
    db = MorkDB()
    db.session.add(EmailStatus(email=email_address, sent_date=datetime.now()))
    db.session.commit()
    db.session.close()



"""Tests of the edx models."""

from mork.edx.factories import EdxUserFactory


def test_edx_models_user_safe_dict(edx_db):
    """Test the `safe_dict` method for the User model."""
    edx_user = EdxUserFactory()

    assert edx_user.safe_dict() == {
        "id": edx_user.id,
        "is_staff": edx_user.is_staff,
        "is_superuser": edx_user.is_superuser,
        "last_login": edx_user.last_login,
    }



"""Mork edx database connection."""

import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from mork.conf import settings

logger = logging.getLogger(__name__)


class OpenEdxDB:
    """Class to connect to the Open edX database."""

    session = None

    def __init__(self, engine=None, session=None):
        """Initialize SqlAlchemy engine and session."""
        if engine is not None:
            self.engine = engine
        else:
            self.engine = create_engine(
                settings.EDX_DB_URL, echo=settings.EDX_DB_DEBUG, pool_pre_ping=True
            )
        if session is not None:
            self.session = session
        else:
            self.session = Session(self.engine)



"""API routes related to tasks."""

import logging

from celery.result import AsyncResult
from fastapi import APIRouter, Depends, Response, status

from mork.api.auth import authenticate_api_key
from mork.api.models import (
    TASK_TYPE_TO_FUNC,
    TaskCreate,
    TaskResponse,
    TaskStatus,
    TaskType,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/tasks", dependencies=[Depends(authenticate_api_key)])


@router.post("/", status_code=status.HTTP_202_ACCEPTED)
async def create_task(task_in: TaskCreate, response: Response) -> TaskResponse:
    """Create a new task."""
    celery_task = TASK_TYPE_TO_FUNC[task_in.type]
    result = celery_task.delay()

    task = TaskResponse(id=result.task_id, status=TaskStatus.PENDING)
    response.headers["location"] = router.url_path_for(
        "get_task_status", **{"task_id": task.id}
    )
    return task


@router.options("/")
async def get_available_tasks(response: Response) -> dict:
    """Get available tasks that can be created."""
    response.headers["allow"] = "POST"
    return {"task_types": list(TaskType)}


@router.get("/status/{task_id}")
async def get_task_status(task_id: str) -> TaskResponse:
    """Get the task status for `task_id`."""
    status = AsyncResult(task_id).state

    return TaskResponse(id=task_id, status=status)



"""Module for defining sessions classes."""

from datetime import datetime
from typing import Optional

from sqlalchemy import distinct, select
from sqlalchemy.orm import Session, load_only
from sqlalchemy.sql.functions import count


class UserMixin:
    """Mixin that adds convenience methods for the User operations."""

    @classmethod
    def get_inactive_users_count(
        cls,
        session: Session,
        threshold_date: datetime,
    ):
        """Get inactive users count from edx database.

        SELECT count(auth_user.id) FROM auth_user
        """
        query = (
            select(count(distinct(cls.id)))
            .prefix_with("SQL_NO_CACHE", dialect="mysql")
            .filter(cls.last_login < threshold_date)
        )
        return session.execute(query).scalar()

    @classmethod
    def get_inactive_users(
        cls,
        session: Session,
        threshold_date: datetime,
        offset: Optional[int] = 0,
        limit: Optional[int] = 0,
    ):
        """Get users from edx database who have not logged in for a specified period.

        SELECT auth_user.id,
            auth_user.username,
            auth_user.email,
            auth_user.is_staff,
            auth_user.is_superuser,
            auth_user.last_login,
        FROM auth_user LIMIT :param_1 OFFSET :param_2
        """
        query = (
            select(cls)
            .prefix_with("SQL_NO_CACHE", dialect="mysql")
            .options(
                load_only(
                    cls.id,
                    cls.username,
                    cls.email,
                    cls.is_staff,
                    cls.is_superuser,
                    cls.last_login,
                ),
            )
            .filter(cls.last_login < threshold_date)
            .offset(offset)
            .limit(limit)
        )
        return session.scalars(query).unique().all()



"""Tests of the mixins class."""

from datetime import datetime, timedelta

from faker import Faker

from mork.edx.factories import EdxUserFactory
from mork.edx.models import User


def test_edx_usermixin_get_inactive_users_count(edx_db):
    """Test the `get_inactive_users_count` method."""
    # 3 users that did not log in for 3 years
    EdxUserFactory.create_batch(3, last_login=Faker().date_time_between(end_date="-3y"))
    # 4 users that logged in recently
    EdxUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get count of users inactive for more than 3 years
    users_count = User.get_inactive_users_count(edx_db.session, threshold_date)

    assert users_count == 3


def test_edx_usermixin_get_inactive_users_count_empty(edx_db):
    """Test the `get_inactive_users_count` method with no inactive users."""
    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get count of users inactive for more than 3 years
    users_count = User.get_inactive_users_count(edx_db.session, threshold_date)

    assert users_count == 0


def test_edx_usermixin_get_inactive_users(edx_db):
    """Test the `get_inactive_users` method."""

    # 3 users that did not log in for 3 years
    inactive_users = EdxUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get all users inactive for more than 3 years
    users = User.get_inactive_users(edx_db.session, threshold_date, offset=0, limit=9)

    assert len(users) == 3
    assert users == inactive_users


def test_edx_usermixin_get_inactive_users_empty(edx_db):
    """Test the `get_inactive_users` method with no inactive users."""

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    users = User.get_inactive_users(edx_db.session, threshold_date, offset=0, limit=9)

    assert users == []


def test_edx_usermixin_get_inactive_users_slice(edx_db):
    """Test the `get_inactive_users` method with a slice."""
    # 3 users that did not log in for 3 years
    inactive_users = EdxUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get two users inactive for more than 3 years
    users = User.get_inactive_users(edx_db.session, threshold_date, offset=0, limit=2)

    assert len(users) == 2
    assert users == inactive_users[:2]


def test_edx_usermixin_get_inactive_users_slice_empty(edx_db):
    """Test the `get_inactive_users` method with an empty slice ."""
    # 3 users that did not log in for 3 years
    EdxUserFactory.create_batch(3, last_login=Faker().date_time_between(end_date="-3y"))
    # 4 users that logged in recently
    EdxUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    users = User.get_inactive_users(edx_db.session, threshold_date, offset=4, limit=9)

    assert users == []



"""Tests for the health check endpoints."""

import pytest


@pytest.mark.anyio
async def test_api_health_lbheartbeat(http_client):
    """Test the load balancer heartbeat healthcheck."""
    response = await http_client.get("/__lbheartbeat__")
    assert response.status_code == 200
    assert response.json() is None



"""Edx database test fixtures."""

import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from mork.conf import settings
from mork.edx.database import OpenEdxDB
from mork.edx.factories import engine, session
from mork.edx.models import Base as EdxBase
from mork.models import Base


@pytest.fixture
def edx_db():
    """"""
    db = OpenEdxDB(engine, session)
    EdxBase.metadata.create_all(engine)
    yield db
    db.session.rollback()
    EdxBase.metadata.drop_all(engine)


@pytest.fixture(scope="session")
def db_engine():
    """Test database engine fixture."""
    engine = create_engine(settings.TEST_DB_URL, echo=False)
    # Create database and tables
    Base.metadata.create_all(engine)

    yield engine

    Base.metadata.drop_all(engine)
    engine.dispose()


@pytest.fixture(scope="function")
def db_session(db_engine):
    """Test session fixture."""
    # Setup
    #
    # Connect to the database and create a non-ORM transaction. Our connection
    # is bound to the test session.
    connection = db_engine.connect()
    transaction = connection.begin()
    session = Session(bind=connection)

    yield session

    # Teardown
    #
    # Rollback everything that happened with the Session above (including
    # explicit commits).
    session.close()
    transaction.rollback()
    connection.close()



"""Mork edx models."""

import datetime
from typing import Optional

from sqlalchemy import DateTime, Index, String
from sqlalchemy.dialects.mysql import INTEGER
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from mork.edx.mixins import UserMixin


class Base(DeclarativeBase):
    """Base class for all models in the database."""

    filtered_attrs = []

    def safe_dict(self):
        """Return a dictionary representation of the model."""
        return {
            c.name: getattr(self, c.name)
            for c in self.__table__.columns
            if c.name not in self.filtered_attrs
        }


class User(UserMixin, Base):
    """Model for the `auth_user` table."""

    __tablename__ = "auth_user"
    __table_args__table_args__ = (
        Index("email", "email", unique=True),
        Index("username", "username", unique=True),
    )
    filtered_attrs = ["username", "email"]

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    username: Mapped[str] = mapped_column(String(30))
    email: Mapped[str] = mapped_column(String(254))
    is_staff: Mapped[int] = mapped_column(INTEGER(1))
    is_superuser: Mapped[int] = mapped_column(INTEGER(1))
    last_login: Mapped[Optional[datetime.datetime]] = mapped_column(DateTime)



"""Main module for Mork API authentication."""

from fastapi import HTTPException, Security, status
from fastapi.security import APIKeyHeader

from mork.conf import settings

api_key_header = APIKeyHeader(name="X-API-Key")


def authenticate_api_key(api_key_header: str = Security(api_key_header)):
    """Authenticate user with any allowed method, using credentials in the header."""
    if api_key_header not in settings.API_KEYS:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Missing or invalid API key",
        )



"""Tests for utility functions."""

from mork.tests.conftest import TEST_STATIC_PATH
from mork.utils import svg_to_datauri


def test_utils_svg_to_datauri_path():
    """Image to base64 from path."""

    red_square_base64 = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCAxIDEiPgo8cGF0aCBkPSJNMCwwaDF2MUgwIiBmaWxsPSIjZjAwIi8+Cjwvc3ZnPg=="  # noqa: E501

    assert (
        svg_to_datauri(TEST_STATIC_PATH / "images/red-square.svg") == red_square_base64
    )



"""Module py.test fixtures."""

# ruff: noqa: F401

from pathlib import Path

from .fixtures.app import http_client
from .fixtures.asynchronous import anyio_backend
from .fixtures.auth import auth_headers
from .fixtures.db import db_engine, db_session, edx_db

TEST_STATIC_PATH = Path(__file__).parent / "static"



"""Tests for Mork extra Jinja tags."""

from mork.templatetags.extra_tags import SVGStaticTag
from mork.tests.conftest import TEST_STATIC_PATH


def test_svgstatictag_render(monkeypatch):
    """Test the SVGStaticTag `render` method returns file encoded in base64."""
    static_filepath = TEST_STATIC_PATH / "images/red-square.svg"

    red_square_base64 = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCAxIDEiPgo8cGF0aCBkPSJNMCwwaDF2MUgwIiBmaWxsPSIjZjAwIi8+Cjwvc3ZnPg=="  # noqa: E501

    monkeypatch.setattr(
        "mork.templatetags.extra_tags.SVGStaticTag.__init__", lambda x: None
    )

    assert SVGStaticTag().render(static_filepath) == red_square_base64


def test_base64tag_render_unknown_file(monkeypatch):
    """Test that the SVGStaticTag `render` method should return an empty string."""
    static_filepath = "unknown-static-file.txt"

    monkeypatch.setattr(
        "mork.templatetags.extra_tags.SVGStaticTag.__init__", lambda x: None
    )

    assert SVGStaticTag().render(static_filepath) == ""
